from gi.repository import GLib
from xml.etree import ElementTree as ET
from .auto_names import *

from .proxy_method import ProxyMethod
from .proxy_property import ProxyProperty
from .proxy_signal import ProxySignal, OnSignal
from .timeout import timeout_to_glib

class ProxyMixin(object):
	__slots__ = ()

	def get(self, bus_name, object_path=None, **kwargs):
		"""Get a remote object.

		Parameters
		----------
		bus_name : string
			Name of the service that exposes this object.
			You may start with "." - then org.freedesktop will be automatically prepended.
		object_path : string, optional
			Path of the object. If not provided, bus_name translated to path format is used.

		Returns
		-------
		ProxyObject implementing all the Interfaces exposed by the remote object.
		Note that it inherits from multiple Interfaces, so the method you want to use
		may be shadowed by another one, eg. from a newer version of the interface.
		Therefore, to interact with only a single interface, use:
		>>> bus.get("org.freedesktop.systemd1")["org.freedesktop.systemd1.Manager"]
		or simply
		>>> bus.get(".systemd1")[".Manager"]
		which will give you access to the one specific interface.
		"""
		# Python 2 sux
		for kwarg in kwargs:
			if kwarg not in ("timeout",):
				raise TypeError(self.__qualname__ + " got an unexpected keyword argument '{}'".format(kwarg))
		timeout = kwargs.get("timeout", None)

		bus_name = auto_bus_name(bus_name)
		object_path = auto_object_path(bus_name, object_path)

		ret = self.con.call_sync(
			bus_name, object_path,
			'org.freedesktop.DBus.Introspectable', "Introspect", None, GLib.VariantType.new("(s)"),
			0, timeout_to_glib(timeout), None)

		if not ret:
			raise KeyError("no such object; you might need to pass object path as the 2nd argument for get()")

		xml, = ret.unpack()

		try:
			introspection = ET.fromstring(xml)
		except:
			raise KeyError("object provides invalid introspection XML")

		return CompositeInterface(introspection)(self, bus_name, object_path)

class ProxyObject(object):
	def __init__(self, bus, bus_name, path, object=None):
		self._bus = bus
		self._bus_name = bus_name
		self._path = path
		self._object = object if object else self

def Interface(iface):

	class interface(ProxyObject):
		@staticmethod
		def _Introspect():
			print(iface.attrib["name"] + ":")
			for member in iface:
				print("\t" + member.tag + " " + member.attrib["name"])
			print()

	interface.__qualname__ = interface.__name__ = iface.attrib["name"]
	interface.__module__ = "DBUS"

	for member in iface:
		member_name = member.attrib["name"]
		if member.tag == "method":
			setattr(interface, member_name, ProxyMethod(interface.__name__, member))
		elif member.tag == "property":
			setattr(interface, member_name, ProxyProperty(interface.__name__, member))
		elif member.tag == "signal":
			signal = ProxySignal(interface.__name__, member)
			setattr(interface, member_name, signal)
			setattr(interface, "on" + member_name, OnSignal(signal))

	return interface

def CompositeInterface(introspection):
	class CompositeObject(ProxyObject):
		def __getitem__(self, iface):
			if iface == "" or iface[0] == ".":
				iface = self._path.replace("/", ".")[1:] + iface
			matching_bases = [base for base in type(self).__bases__ if base.__name__ == iface]

			if len(matching_bases) == 0:
				raise KeyError(iface)
			assert(len(matching_bases) == 1)

			iface_class = matching_bases[0]
			return iface_class(self._bus, self._bus_name, self._path, self)

		@classmethod
		def _Introspect(cls):
			for iface in cls.__bases__:
				try:
					iface._Introspect()
				except:
					pass

	ifaces = sorted([x for x in introspection if x.tag == "interface"], key=lambda x: int(x.attrib["name"].startswith("org.freedesktop.DBus.")))
	if not ifaces:
		raise KeyError("object does not export any interfaces; you might need to pass object path as the 2nd argument for get()")
	CompositeObject.__bases__ = tuple(Interface(iface) for iface in ifaces)
	CompositeObject.__name__ = "<CompositeObject>"
	CompositeObject.__qualname__ = "<CompositeObject>(" + "+".join(x.__name__ for x in CompositeObject.__bases__) + ")"
	CompositeObject.__module__ = "DBUS"
	return CompositeObject



from gi.repository import GLib, GObject

def timeout_to_glib(timeout):
	if timeout is None:
		try:
			return GLib.MAXINT
		except AttributeError:
			# GLib < 2.46
			return GObject.G_MAXINT
	else:
		try:
			timeout = timeout.total_seconds()
		except AttributeError:
			pass
		return int(timeout * 1000)



from .exitable import ExitableWithAliases

class NameOwner(ExitableWithAliases("unown")):
	__slots__ = ()

	def __init__(self, bus, name, allow_replacement, replace):
		flags = 4 | (1 if allow_replacement else 0) | (2 if replace else 0)
		res = bus.dbus.RequestName(name, flags)
		if res == 1:
			self._at_exit(lambda: bus.dbus.ReleaseName(name))
			return # OK
		if res == 3:
			raise RuntimeError("name already exists on the bus")
		if res == 4:
			raise RuntimeError("you're already the owner of this name")
		raise RuntimeError("cannot take ownership of the name")

class RequestNameMixin(object):
	__slots__ = ()

	def request_name(self, name, allow_replacement=True, replace=False):
		"""Aquires a bus name.

		Returns
		-------
		NameOwner
			An object you can use as a context manager to unown the name later.
		"""
		return NameOwner(self, name, allow_replacement, replace)



from gi.repository import Gio
from .exitable import ExitableWithAliases
import warnings

class NameOwner(ExitableWithAliases("unown")):
	Flags = Gio.BusNameOwnerFlags
	__slots__ = ()

	def __init__(self, con, name, flags, name_aquired_handler, name_lost_handler):
		id = Gio.bus_own_name_on_connection(con, name, flags, name_aquired_handler, name_lost_handler)
		self._at_exit(lambda: Gio.bus_unown_name(id))

class NameWatcher(ExitableWithAliases("unwatch")):
	Flags = Gio.BusNameWatcherFlags
	__slots__ = ()

	def __init__(self, con, name, flags, name_appeared_handler, name_vanished_handler):
		id = Gio.bus_watch_name_on_connection(con, name, flags, name_appeared_handler, name_vanished_handler)
		self._at_exit(lambda: Gio.bus_unwatch_name(id))

class OwnMixin(object):
	__slots__ = ()
	NameOwnerFlags = NameOwner.Flags

	def own_name(self, name, flags=0, name_aquired=None, name_lost=None):
		"""[DEPRECATED] Asynchronously aquires a bus name.

		Starts acquiring name on the bus specified by bus_type and calls
		name_acquired and name_lost when the name is acquired respectively lost.

		To receive name_aquired and name_lost callbacks, you need an event loop.
		https://github.com/LEW21/pydbus/blob/master/doc/tutorial.rst#setting-up-an-event-loop

		Parameters
		----------
		name : string
			Bus name to aquire
		flags : NameOwnerFlags, optional
		name_aquired : callable, optional
			Invoked when name is acquired
		name_lost : callable, optional
			Invoked when name is lost

		Returns
		-------
		NameOwner
			An object you can use as a context manager to unown the name later.

		See Also
		--------
		See https://developer.gnome.org/gio/2.44/gio-Owning-Bus-Names.html#g-bus-own-name
		for more information.
		"""
		warnings.warn("own_name() is deprecated, use request_name() instead.", DeprecationWarning)

		name_aquired_handler = (lambda con, name: name_aquired()) if name_aquired is not None else None
		name_lost_handler    = (lambda con, name: name_lost())    if name_lost    is not None else None
		return NameOwner(self.con, name, flags, name_aquired_handler, name_lost_handler)

class WatchMixin(object):
	__slots__ = ()
	NameWatcherFlags = NameWatcher.Flags

	def watch_name(self, name, flags=0, name_appeared=None, name_vanished=None):
		"""Asynchronously watches a bus name.

		Starts watching name on the bus specified by bus_type and calls
		name_appeared and name_vanished when the name is known to have a owner
		respectively known to lose its owner.

		To receive name_appeared and name_vanished callbacks, you need an event loop.
		https://github.com/LEW21/pydbus/blob/master/doc/tutorial.rst#setting-up-an-event-loop

		Parameters
		----------
		name : string
			Bus name to watch
		flags : NameWatcherFlags, optional
		name_appeared : callable, optional
			Invoked when name is known to exist
			Called as name_appeared(name_owner).
		name_vanished : callable, optional
			Invoked when name is known to not exist

		Returns
		-------
		NameWatcher
			An object you can use as a context manager to unwatch the name later.

		See Also
		--------
		See https://developer.gnome.org/gio/2.44/gio-Watching-Bus-Names.html#g-bus-watch-name
		for more information.
		"""
		name_appeared_handler = (lambda con, name, name_owner: name_appeared(name_owner)) if name_appeared is not None else None
		name_vanished_handler = (lambda con, name:             name_vanished())           if name_vanished is not None else None
		return NameWatcher(self.con, name, flags, name_appeared_handler, name_vanished_handler)
